//
// Map Demo
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

// *** Required GameX stuff:

#define GAMEX_MAIN
#pragma comment (lib, "gamex.lib")
#include "gamex.hpp"

// *** Global constants and variables:

const int NUM_TILE_IMAGES = 4;  // number of tile texture pictures we currently have
const int MAX_MAP_EXTENT = 128; // maximum width or height a map file is allowed to be

ImageX characterImage;						// character image (just 1 for now)
ImageX tile [NUM_TILE_IMAGES];				// tile images
CameraX cam;

int map [MAX_MAP_EXTENT][MAX_MAP_EXTENT];	// map texture values
int mapBright [MAX_MAP_EXTENT][MAX_MAP_EXTENT];	// map brightness values 
int currentMap=0, mapWidth=0, mapHeight=0;  // map attributes
float charX, charY; // position of character, in tile coordinates
int charFrame = 0;
bool viewIs2D = true;
int gameHasStarted = false;

// *** Helper functions:

bool loadMap (int mapNum);
bool saveMap (void);
void updateIntro (void);
void updateCharacter (void);
void doMiscInput (void);
void drawHelpText (void);
void startDraw (void);
void drawTiles (void);
void drawCharacter (void);

// Required GameX initialize function (called when program starts)
// this function should initialize GameX and prepare anything else we need
void GameInit (void)
{
	// Initialize GameX:

	GameX.Initialize ("Map Demo", VIDEO_16BIT|RUN_AUTOSHOWINFO, 800, 600);

	// Load the tile images:

	char fileName [MAX_PATH];
	for(int i = 0 ; i < NUM_TILE_IMAGES ; i++) {
		sprintf(fileName, "Images\\tile%d.bmp", i); // note: tile0.bmp comes first, and tile10.bmp comes after tile9.bmp
		tile[i].Load(fileName);
	}

	// Load the character images:

	characterImage.Load("Images\\character.bmp", true); // with transparency (mask)

	// Set up the camera: (useful even for 2D drawing because it calculates offsets for us)

	cam.SetWindow(0,0,GameX.GetWidth(),GameX.GetHeight()); // tell the camera to project to the whole screen
	cam.SetUpVector(Vector3DF(0,-1.0f,0)); // set -y to "up" so it's like normal 2D coordinates
	GameX.SetDrawCam(&cam);
}

// Required GameX main function (called repeatedly while game is running)
// this function should update the game state but should NOT draw that state
void GameRun (void)
{
	if(gameHasStarted) {
		updateCharacter();
		doMiscInput();
	} else {
		updateIntro();
	}
}

// Required GameX draw function (called repeatedly while game is running)
// this function should draw the game state and do as little else as possible
void GameDraw (void)
{
	if(gameHasStarted) {
		startDraw();
		drawTiles();
		drawCharacter();
	} else {
		drawHelpText();
	}
}

// Loads a map file of the given number, return true if succeeded, false if failed
bool loadMap (int mapNum)
{
	// Open the map file and prepare to read from it:
	
	char fileName [MAX_PATH];
	sprintf(fileName, "Maps\\map%d.txt", mapNum); // (we could use .map or something instead of .txt)
	FILE* file = fopen(fileName, "r");
	if(file == NULL) {
		char str [64]; sprintf(str, "Failed to load map: %s", fileName);
		GameX.ErrorDialog(str); // (error-checking is good for your computer and your sanity)
		return false;
	}

	// Now we're going to read the map in using fscanf to scan for the data.
	// Any plain text in the scan string is something that must be next in the file, but doesn't change anything.
	// %d means we read in an integer, into an int which we give a pointer to as an extra argument.
	// %f means we read in a floating point value, into a float or double we give a pointer to as an extra argument.
	// %s means we read in a string (until whitespace), into a C-string we give as an extra argument
	// \"%[^\"]\" means we read in a file pathname string (surrounded by quotes), into a C-string we give

	fscanf(file, " width: %d height: %d", &mapWidth, &mapHeight);

	int i, j;

	fscanf(file, " texture data:");
	for(i = 0 ; i < mapHeight ; i++)
		for(j = 0 ; j < mapWidth ; j++)
			fscanf(file, " %d", &map[i][j]);

	fscanf(file, " brightness data:");
	for(i = 0 ; i < mapHeight ; i++)
		for(j = 0 ; j < mapWidth ; j++)
			fscanf(file, " %d", &mapBright[i][j]), mapBright[i][j] *= 16;

	fclose(file); // done loading from file

	currentMap = mapNum;
	return true;
}

// Loads the current map file to the next available number, return true if succeeded, false if cancelled or failed
bool saveMap (void)
{
	// Find the first map#.txt filename that isn't already taken:

	File file;
	char fileName [MAX_PATH];
	int num = 0;
	do {
		sprintf(fileName,"Maps\\map%d.txt",num);
		num++;
	} while (file.Exist(fileName) == FILE_EXIST_YES);

	char buf [128];

	// Ask the user if they really want to save the file:
	
	sprintf(buf, "Save this map to \"%s\"?", fileName);
	int value = GameX.ChoiceDialog(buf, false, 1);
	if(value == 0) return false;

	// If they said Yes, create and open the file for writing:

	int successCode = file.Open(fileName,FILE_WRITE | FILE_SEQUENTIAL | FILE_AUTOCREATE);
	if(successCode != FILE_STATUS_OK) {
		char str [64]; sprintf(str, "Failed to save map: %s", fileName);
		GameX.ErrorDialog(str);
		return false;
	}

	// Now we begin writing to the map file,
	// writing out exactly what loadMap reads in:

	sprintf(buf, "width: %d \nheight: %d \n", mapWidth, mapHeight);
	file.WriteS(buf);

	int i, j;

	file.WriteS("\ntexture data: \n");
	for(i = 0 ; i < mapHeight ; i++) {
		for(j = 0 ; j < mapWidth ; j++) {
			sprintf(buf, "%02d ", map[i][j]);
			file.WriteS(buf);
		}
		file.WriteS("\n"); // insert newline after each row (not required, just makes it more readable)
	}
	file.WriteS("\nbrightness data: \n");
	for(i = 0 ; i < mapHeight ; i++) {
		for(j = 0 ; j < mapWidth ; j++) {
			sprintf(buf, "%d ", mapBright[i][j]/16);
			file.WriteS(buf);
		}
		file.WriteS("\n");
	}

	file.Close(); // done writing to file

	return true;
}

// Handles input at the intro screen
void updateIntro (void)
{
	if(GameX.GetKeyPress(KEY_LEFT) || GameX.GetKeyPress(KEY_RIGHT)) {
		bool done = false;
		int times = 0; // (infinite loop prevention, in case all map files are gone)
		int newNum = currentMap;
		while(!done && times < 1001) {
			if(GameX.GetKeyPress(KEY_LEFT))  newNum--;
	   else if(GameX.GetKeyPress(KEY_RIGHT)) newNum++;
			char fileName [MAX_PATH];
			sprintf(fileName, "Maps\\map%d.txt", newNum); // (we could use .map or something instead of .txt)
			File test;
			if(test.Exist(fileName)) {
				currentMap = newNum;
				done = true;
			}
			times++;
			if(newNum > 999) newNum = -1;
			else if(newNum < 0) newNum = 1000;
				
		}
	}
	if(GameX.GetKeyPress(KEY_ENTER) || GameX.GetKeyPress(KEY_SPACE)) {

		// Load the map from file:
		bool success = loadMap(currentMap);
		if(success) {
			// Set up the character:
			charX = (float)mapWidth/2.0f;  // start the character centered
			charY = (float)mapHeight/2.0f; // in the middle of the map

			gameHasStarted = true; // begin the game
		}
	}
}

// Handles character movement and input
void updateCharacter (void)
{
	float xPlus = 0.0f, yPlus = 0.0f;
	if(GameX.IsKeyDown(KEY_LEFT))
		xPlus -= 0.1f;
	if(GameX.IsKeyDown(KEY_RIGHT))
		xPlus += 0.1f;
	if(GameX.IsKeyDown(KEY_UP))
		yPlus -= 0.1f;
	if(GameX.IsKeyDown(KEY_DOWN))
		yPlus += 0.1f;

	if(xPlus || yPlus) {

		// update position by movement amount:
		charX += xPlus, charY += yPlus;

		// stop character from going out of map boundaries:
		if(charX < 0.0f) charX = 0.0f;
		if(charY < 0.0f) charY = 0.0f;
		if(charX > (float)mapWidth -0.001f) charX = (float)mapWidth -0.001f;
		if(charY > (float)mapHeight-0.001f) charY = (float)mapHeight-0.001f;

		if(fabsf(xPlus) < fabsf(yPlus)) { // if movement is more vertical than sideways
			if(yPlus < 0) { // movement is up
				charFrame = 9;
			} else { // movement is down
				charFrame = 6;
			}
		} else { // if movement is more (or equally) sideways than vertical
			if(xPlus < 0) { // movement is left
				charFrame = 0;
			} else { // movement is right
				charFrame = 3;
			}
		}

		// animate:
		const int walkPat [6] = {0,1,1,0,2,2}; // straight, right foot forward, straight, left foot forward, repeat
		charFrame += walkPat[(GameX.GetGlobalCounter()/8)%6]; // 8 is walking slowness, 6 is array length of walkPat

	} else {
		// revert to 1st frame of animation
		charFrame -= charFrame%3;
	}
}

// Handles miscellanous input like editing the map and changing the view
void doMiscInput (void)
{
	// Map editing input:
	// NOTE: A real map editor would need better controls and more features.
	//       (i.e. a way to change the map size, a way to lay down large amounts of one texture,
	//        a way to select where to save, a finer level of control over brightness, etc.)

	if(GameX.GetKeyPress(KEY_T)) {
		map[(int)charY][(int)charX] ++;
		if(map[(int)charY][(int)charX] >= NUM_TILE_IMAGES)
			map[(int)charY][(int)charX] = -1;
	}

	if(GameX.GetKeyPress(KEY_B)) {
		mapBright[(int)charY][(int)charX] += 16;
		mapBright[(int)charY][(int)charX] %= 256;
	}

	if(GameX.GetKeyPress(KEY_S))
		saveMap();

	if(GameX.GetKeyPress(KEY_L) || GameX.GetKeyPress(KEY_Q))
		gameHasStarted = false; // returns to the intro screen so you can load another map

	if(viewIs2D) {
		if(GameX.IsKeyDown(KEY_3) || GameX.GetKeyPress(KEY_SPACE)) {
			viewIs2D = false; // set view to 3D mode
			cam.SetUpVector(Vector3DF(0,0,-1.0f)); // set -z to "up"
		}
	} else {
		if(GameX.IsKeyDown(KEY_2) || GameX.GetKeyPress(KEY_SPACE)) {
			viewIs2D = true; // set view to 2D mode
			cam.SetUpVector(Vector3DF(0,-1.0f,0)); // set -y to "up" so it's like normal 2D coordinates
		}
	}
}

// Draws the text at the intro screen
void drawHelpText (void)
{
	GameX.ClearScreen(); // black background
	GameX.DrawText(20,20,"Controls:");
	GameX.DrawText(40,40,"The arrow keys move your character.",128,255,255);
	GameX.DrawText(40,60,"The T key changes the texture.",128,255,255);
	GameX.DrawText(40,80,"The B key changes the brightness.",128,255,255);
	GameX.DrawText(40,100,"The S key saves the map to file.",128,255,255);
	GameX.DrawText(40,120,"The L key loads a map from file.",128,255,255);
	GameX.DrawText(40,140,"Space, 2, or 3 switch the viewing mode.",128,255,255);
	GameX.DrawText(20,180,"Press the <- or -> arrow keys now to select which map file to load,");
	GameX.DrawText(20,200,"Then press Enter or the Space Bar to continue...");

	char str [128];
	sprintf(str,"Selected map:  Maps\\map%d.txt",currentMap);
	GameX.DrawText(40,240,str, 255,255,0);
}

// Begins drawing main field scene
void startDraw (void)
{
	// draw sky background:
	GameX.FillScreen(ColorX(128,128,192,255, 128,128,192,255, 128,255,192,255, 128,255,192,255));

	// and update the camera, in case the character moved:
	if(viewIs2D) {
		cam.SetFromPosition(Vector3DF(charX,charY,-10)); // position camera is at (the -10 means 10 tiles fit across the screen)
		cam.SetToPosition(Vector3DF(charX,charY,0)); // position camera is looking toward
	} else {
		float xCam = max(4,min(mapWidth-4,charX)) + 0; // same as char position except near an edge, then it stays 4 away from edge
		float yCam = max(4,min(mapHeight-4,charY)) + 8; // the numbers added at the end determines the angle of viewing
		cam.SetFromPosition(Vector3DF(xCam,yCam,-8)); // position camera is at
		cam.SetToPosition(Vector3DF(charX,charY,0)); // position camera is looking toward
	}
}

// Draws map field tiles
void drawTiles (void)
{
	// loop through the rows and columns, drawing each square tile of the map:
	for(int i = 0 ; i < mapHeight ; i++) {
		for(int j = 0 ; j < mapWidth ; j++) {
			int index = map[i][j];
			if(index >= 0 && index < NUM_TILE_IMAGES) {

				int bright = mapBright[i][j];
				GameX.SetDrawMode(DRAW_PLAIN | DRAWOP_BRIGHT);
				GameX.SetDrawShading(ColorX(bright,bright,bright)); // shade whole tile with that brightness (but a simple change can smooth it out...)

				float x = (float)j;
				float y = (float)i;
				GameX.DrawTexturedPolygon3D(&tile[index], Vector3DF(x,y,0),Vector3DF(x+1,y,0),Vector3DF(x,y+1,0),Vector3DF(x+1,y+1,0));
			}
		}
	}
}

// Draws the character at current position
void drawCharacter (void)
{
	int left = (charFrame % 4) * 32;
	int top = (charFrame / 4) * 32;
	GameX.SetDrawPart(left,top,left+32,top+32);

	int bright = mapBright[(int)charY][(int)charX];
	GameX.SetDrawShading(ColorX(bright,bright,bright));

	GameX.SetDrawMode(DRAW_NORMAL | DRAWOP_CENTERED | DRAWOP_BRIGHT);
	GameX.DrawImage3D(&characterImage, Vector3DF(charX,charY,-0.75f));
}
